@@ -13,17 +13,17 @@
 %                      Render Text Onto A Canvas Image.                       %
 %                                                                             %
 %                              Software Design                                %
-%                                John Cristy                                  %
+%                                   Cristy                                    %
 %                                 July 1992                                   %
 %                                                                             %
 %                                                                             %
-%  Copyright 1999-2013 ImageMagick Studio LLC, a non-profit organization      %
+%  Copyright 1999-2020 ImageMagick Studio LLC, a non-profit organization      %
 %  dedicated to making software imaging solutions freely available.           %
 %                                                                             %
 %  You may not use this file except in compliance with the License.  You may  %
 %  obtain a copy of the License at                                            %
 %                                                                             %
-%    http://www.imagemagick.org/script/license.php                            %
+%    https://imagemagick.org/script/license.php                               %
 %                                                                             %
 %  Unless required by applicable law or agreed to in writing, software        %
 %  distributed under the License is distributed on an "AS IS" BASIS,          %
@@ -119,7 +119,7 @@ static MagickBooleanType IsTXT(const unsigned char *magick,const size_t length)
     return(MagickFalse);
   if (LocaleNCompare((const char *) magick,MagickID,strlen(MagickID)) != 0)
     return(MagickFalse);
-  count=(ssize_t) sscanf((const char *) magick+32,"%lu,%lu,%lu,%s",&columns,
+  count=(ssize_t) sscanf((const char *) magick+32,"%lu,%lu,%lu,%32s",&columns,
     &rows,&depth,colorspace);
   if (count != 4)
     return(MagickFalse);
@@ -157,18 +157,20 @@ static MagickBooleanType IsTXT(const unsigned char *magick,const size_t length)
 %    o exception: return any errors or warnings in this structure.
 %
 */
-static Image *ReadTEXTImage(const ImageInfo *image_info,Image *image,
-  char *text,ExceptionInfo *exception)
+static Image *ReadTEXTImage(const ImageInfo *image_info,
+  ExceptionInfo *exception)
 {
   char
     filename[MaxTextExtent],
     geometry[MaxTextExtent],
-    *p;
+    *p,
+    text[MaxTextExtent];
 
   DrawInfo
     *draw_info;
 
   Image
+    *image,
     *texture;
 
   MagickBooleanType
@@ -190,12 +192,21 @@ static Image *ReadTEXTImage(const ImageInfo *image_info,Image *image,
     Open image file.
   */
   assert(image_info != (const ImageInfo *) NULL);
-  assert(image_info->signature == MagickSignature);
+  assert(image_info->signature == MagickCoreSignature);
   if (image_info->debug != MagickFalse)
     (void) LogMagickEvent(TraceEvent,GetMagickModule(),"%s",
       image_info->filename);
   assert(exception != (ExceptionInfo *) NULL);
-  assert(exception->signature == MagickSignature);
+  assert(exception->signature == MagickCoreSignature);
+  image=AcquireImage(image_info);
+  status=OpenBlob(image_info,image,ReadBinaryBlobMode,exception);
+  if (status == MagickFalse)
+    {
+      image=DestroyImageList(image);
+      return((Image *) NULL);
+    }
+  (void) memset(text,0,sizeof(text));
+  (void) ReadBlobString(image,text);
   /*
     Set the page geometry.
   */
@@ -228,6 +239,14 @@ static Image *ReadTEXTImage(const ImageInfo *image_info,Image *image,
     delta.x)+0.5);
   image->rows=(size_t) floor((((double) page.height*image->y_resolution)/
     delta.y)+0.5);
+  status=SetImageExtent(image,image->columns,image->rows);
+  if (status != MagickFalse)
+    status=ResetImagePixels(image,&image->exception);
+  if (status == MagickFalse)
+    {
+      InheritException(exception,&image->exception);
+      return(DestroyImageList(image));
+    }
   image->page.x=0;
   image->page.y=0;
   texture=(Image *) NULL;
@@ -249,15 +268,18 @@ static Image *ReadTEXTImage(const ImageInfo *image_info,Image *image,
   (void) SetImageBackgroundColor(image);
   draw_info=CloneDrawInfo(image_info,(DrawInfo *) NULL);
   (void) CloneString(&draw_info->text,image_info->filename);
-  (void) FormatLocaleString(geometry,MaxTextExtent,"0x0%+ld%+ld",(long) page.x,
-    (long) page.y);
+  (void) FormatLocaleString(geometry,MaxTextExtent,"%gx%g%+g%+g",(double)
+    image->columns,(double) image->rows,(double) page.x,(double) page.y);
   (void) CloneString(&draw_info->geometry,geometry);
   status=GetTypeMetrics(image,draw_info,&metrics);
   if (status == MagickFalse)
-    ThrowReaderException(TypeError,"UnableToGetTypeMetrics");
+    {
+      draw_info=DestroyDrawInfo(draw_info);
+      ThrowReaderException(TypeError,"UnableToGetTypeMetrics");
+    }
   page.y=(ssize_t) ceil((double) page.y+metrics.ascent-0.5);
-  (void) FormatLocaleString(geometry,MaxTextExtent,"0x0%+ld%+ld",(long) page.x,
-    (long) page.y);
+  (void) FormatLocaleString(geometry,MaxTextExtent,"%gx%g%+g%+g",(double)
+    image->columns,(double) image->rows,(double) page.x,(double) page.y);
   (void) CloneString(&draw_info->geometry,geometry);
   (void) CopyMagickString(filename,image_info->filename,MaxTextExtent);
   if (*draw_info->text != '\0')
@@ -273,7 +295,8 @@ static Image *ReadTEXTImage(const ImageInfo *image_info,Image *image,
     offset+=(ssize_t) (metrics.ascent-metrics.descent);
     if (image->previous == (Image *) NULL)
       {
-        status=SetImageProgress(image,LoadImageTag,offset,image->rows);
+        status=SetImageProgress(image,LoadImageTag,(MagickOffsetType) offset,
+          image->rows);
         if (status == MagickFalse)
           break;
       }
@@ -302,8 +325,8 @@ static Image *ReadTEXTImage(const ImageInfo *image_info,Image *image,
     AcquireNextImage(image_info,image);
     if (GetNextImageInList(image) == (Image *) NULL)
       {
-        image=DestroyImageList(image);
-        return((Image *) NULL);
+        status=MagickFalse;
+        break;
       }
     image->next->columns=image->columns;
     image->next->rows=image->rows;
@@ -330,6 +353,8 @@ static Image *ReadTEXTImage(const ImageInfo *image_info,Image *image,
     texture=DestroyImage(texture);
   draw_info=DestroyDrawInfo(draw_info);
   (void) CloseBlob(image);
+  if (status == MagickFalse)
+    return(DestroyImageList(image));
   return(GetFirstImageInList(image));
 }
 
@@ -365,24 +390,23 @@ static Image *ReadTXTImage(const ImageInfo *image_info,ExceptionInfo *exception)
     colorspace[MaxTextExtent],
     text[MaxTextExtent];
 
+  double
+    max_value,
+    x_offset,
+    y_offset;
+
   Image
     *image;
 
   IndexPacket
     *indexes;
 
-  long
-    type,
-    x_offset,
-    y,
-    y_offset;
-
-  LongPixelPacket
-    pixel;
-
   MagickBooleanType
     status;
 
+  MagickPixelPacket
+    pixel;
+
   QuantumAny
     range;
 
@@ -394,24 +418,25 @@ static Image *ReadTXTImage(const ImageInfo *image_info,ExceptionInfo *exception)
     *q;
 
   ssize_t
-    count;
+    count,
+    type,
+    y;
 
   unsigned long
     depth,
     height,
-    max_value,
     width;
 
   /*
     Open image file.
   */
   assert(image_info != (const ImageInfo *) NULL);
-  assert(image_info->signature == MagickSignature);
+  assert(image_info->signature == MagickCoreSignature);
   if (image_info->debug != MagickFalse)
     (void) LogMagickEvent(TraceEvent,GetMagickModule(),"%s",
       image_info->filename);
   assert(exception != (ExceptionInfo *) NULL);
-  assert(exception->signature == MagickSignature);
+  assert(exception->signature == MagickCoreSignature);
   image=AcquireImage(image_info);
   status=OpenBlob(image_info,image,ReadBinaryBlobMode,exception);
   if (status == MagickFalse)
@@ -419,21 +444,37 @@ static Image *ReadTXTImage(const ImageInfo *image_info,ExceptionInfo *exception)
       image=DestroyImageList(image);
       return((Image *) NULL);
     }
-  (void) ResetMagickMemory(text,0,sizeof(text));
+  (void) memset(text,0,sizeof(text));
   (void) ReadBlobString(image,text);
   if (LocaleNCompare((char *) text,MagickID,strlen(MagickID)) != 0)
-    return(ReadTEXTImage(image_info,image,text,exception));
+    ThrowReaderException(CorruptImageError,"ImproperImageHeader");
+  x_offset=(-1.0);
+  y_offset=(-1.0);
+  q=(PixelPacket *) NULL;
   do
   {
+    width=0;
+    height=0;
+    max_value=0.0;
     *colorspace='\0';
-    count=(ssize_t) sscanf(text+32,"%lu,%lu,%lu,%s",&width,&height,&max_value,
+    count=(ssize_t) sscanf(text+32,"%lu,%lu,%lf,%32s",&width,&height,&max_value,
       colorspace);
-    if (count != 4)
+    if ((count != 4) || (width == 0) || (height == 0) || (max_value == 0.0))
       ThrowReaderException(CorruptImageError,"ImproperImageHeader");
     image->columns=width;
     image->rows=height;
-    for (depth=1; (GetQuantumRange(depth)+1) < max_value; depth++) ;
+    if ((max_value == 0.0) || (max_value > 18446744073709551615.0))
+      ThrowReaderException(CorruptImageError,"ImproperImageHeader");
+    for (depth=1; (GetQuantumRange(depth)+1.0) < max_value; depth++) ;
     image->depth=depth;
+    status=SetImageExtent(image,image->columns,image->rows);
+    if (status != MagickFalse)
+      status=ResetImagePixels(image,&image->exception);
+    if (status == MagickFalse)
+      {
+        InheritException(exception,&image->exception);
+        return(DestroyImageList(image));
+      }
     LocaleLower(colorspace);
     i=(ssize_t) strlen(colorspace)-1;
     image->matte=MagickFalse;
@@ -446,56 +487,129 @@ static Image *ReadTXTImage(const ImageInfo *image_info,ExceptionInfo *exception)
     if (type < 0)
       ThrowReaderException(CorruptImageError,"ImproperImageHeader");
     image->colorspace=(ColorspaceType) type;
-    (void) ResetMagickMemory(&pixel,0,sizeof(pixel));
+    (void) memset(&pixel,0,sizeof(pixel));
     (void) SetImageBackgroundColor(image);
     range=GetQuantumRange(image->depth);
+    status=MagickTrue;
     for (y=0; y < (ssize_t) image->rows; y++)
     {
+      double
+        blue,
+        green,
+        index,
+        opacity,
+        red;
+
+      if (status == MagickFalse)
+        break;
+      red=0.0;
+      green=0.0;
+      blue=0.0;
+      index=0.0;
+      opacity=0.0;
       for (x=0; x < (ssize_t) image->columns; x++)
       {
         if (ReadBlobString(image,text) == (char *) NULL)
-          break;
-        if (image->colorspace == CMYKColorspace)
+          {
+            status=MagickFalse;
+            break;
+          }
+        switch (image->colorspace)
+        {
+          case LinearGRAYColorspace:
+          case GRAYColorspace:
           {
             if (image->matte != MagickFalse)
-              count=(ssize_t) sscanf(text,"%ld,%ld: (%u,%u,%u,%u,%u",&x_offset,
-                &y_offset,&pixel.red,&pixel.green,&pixel.blue,&pixel.index,
-                &pixel.opacity);
-            else
-              count=(ssize_t) sscanf(text,"%ld,%ld: (%u,%u,%u,%u",&x_offset,
-                &y_offset,&pixel.red,&pixel.green,&pixel.blue,&pixel.index);
+              {
+                (void) sscanf(text,"%lf,%lf: (%lf%*[%,]%lf%*[%,]",&x_offset,
+                  &y_offset,&red,&opacity);
+                green=red;
+                blue=red;
+                break;
+              }
+            (void) sscanf(text,"%lf,%lf: (%lf%*[%,]",&x_offset,&y_offset,&red);
+            green=red;
+            blue=red;
+            break;
+          }
+          case CMYKColorspace:
+          {
+            if (image->matte != MagickFalse)
+              {
+                (void) sscanf(text,
+                  "%lf,%lf: (%lf%*[%,]%lf%*[%,]%lf%*[%,]%lf%*[%,]%lf%*[%,]",
+                  &x_offset,&y_offset,&red,&green,&blue,&index,&opacity);
+                break;
+              }
+            (void) sscanf(text,
+              "%lf,%lf: (%lf%*[%,]%lf%*[%,]%lf%*[%,]%lf%*[%,]",&x_offset,
+              &y_offset,&red,&green,&blue,&index);
+            break;
+          }
+          default:
+          {
+            if (image->matte != MagickFalse)
+              {
+                (void) sscanf(text,
+                  "%lf,%lf: (%lf%*[%,]%lf%*[%,]%lf%*[%,]%lf%*[%,]",
+                  &x_offset,&y_offset,&red,&green,&blue,&opacity);
+                break;
+              }
+            (void) sscanf(text,"%lf,%lf: (%lf%*[%,]%lf%*[%,]%lf%*[%,]",
+              &x_offset,&y_offset,&red,&green,&blue);
+            break;
+          }
+        }
+        if (strchr(text,'%') != (char *) NULL)
+          {
+            red*=0.01*range;
+            green*=0.01*range;
+            blue*=0.01*range;
+            index*=0.01*range;
+            opacity*=0.01*range;
           }
-        else
-          if (image->matte != MagickFalse)
-            count=(ssize_t) sscanf(text,"%ld,%ld: (%u,%u,%u,%u",&x_offset,
-              &y_offset,&pixel.red,&pixel.green,&pixel.blue,&pixel.opacity);
-          else
-            count=(ssize_t) sscanf(text,"%ld,%ld: (%u,%u,%u",&x_offset,
-              &y_offset,&pixel.red,&pixel.green,&pixel.blue);
         if (image->colorspace == LabColorspace)
           {
-            pixel.green+=(range+1)/2.0;
-            pixel.blue+=(range+1)/2.0;
+            green+=(range+1)/2.0;
+            blue+=(range+1)/2.0;
           }
-        if (count < 5)
-          continue;
-        q=GetAuthenticPixels(image,x_offset,y_offset,1,1,exception);
+        pixel.red=(MagickRealType) ScaleAnyToQuantum((QuantumAny)
+          MagickMax(red+0.5,0.0),range);
+        pixel.green=(MagickRealType) ScaleAnyToQuantum((QuantumAny)
+          MagickMax(green+0.5,0.0),range);
+        pixel.blue=(MagickRealType) ScaleAnyToQuantum((QuantumAny)
+          MagickMax(blue+0.5,0.0),range);
+        pixel.index=(MagickRealType) ScaleAnyToQuantum((QuantumAny)
+          MagickMax(index+0.5,0.0),range);
+        pixel.opacity=(MagickRealType) ScaleAnyToQuantum((QuantumAny)
+          MagickMax(opacity+0.5,0.0),range);
+        q=GetAuthenticPixels(image,(ssize_t) x_offset,(ssize_t) y_offset,1,1,
+          exception);
         if (q == (PixelPacket *) NULL)
-          continue;
-        SetPixelRed(q,ScaleAnyToQuantum(pixel.red,range));
-        SetPixelGreen(q,ScaleAnyToQuantum(pixel.green,range));
-        SetPixelBlue(q,ScaleAnyToQuantum(pixel.blue,range));
+          {
+            status=MagickFalse;
+            break;
+          }
+        SetPixelRed(q,pixel.red);
+        SetPixelGreen(q,pixel.green);
+        SetPixelBlue(q,pixel.blue);
         if (image->colorspace == CMYKColorspace)
           {
             indexes=GetAuthenticIndexQueue(image);
-            SetPixelIndex(indexes,ScaleAnyToQuantum(pixel.index,range));
+            SetPixelIndex(indexes,pixel.index);
           }
         if (image->matte != MagickFalse)
-          SetPixelAlpha(q,ScaleAnyToQuantum(pixel.opacity,range));
+          SetPixelAlpha(q,pixel.opacity);
         if (SyncAuthenticPixels(image,exception) == MagickFalse)
-          break;
+          {
+            status=MagickFalse;
+            break;
+          }
       }
     }
+    if (status == MagickFalse)
+      break;
+    *text='\0';
     (void) ReadBlobString(image,text);
     if (LocaleNCompare((char *) text,MagickID,strlen(MagickID)) == 0)
       {
@@ -505,8 +619,8 @@ static Image *ReadTXTImage(const ImageInfo *image_info,ExceptionInfo *exception)
         AcquireNextImage(image_info,image);
         if (GetNextImageInList(image) == (Image *) NULL)
           {
-            image=DestroyImageList(image);
-            return((Image *) NULL);
+            status=MagickFalse;
+            break;
           }
         image=SyncNextImageInList(image);
         status=SetImageProgress(image,LoadImagesTag,TellBlob(image),
@@ -516,6 +630,8 @@ static Image *ReadTXTImage(const ImageInfo *image_info,ExceptionInfo *exception)
       }
   } while (LocaleNCompare((char *) text,MagickID,strlen(MagickID)) == 0);
   (void) CloseBlob(image);
+  if (q == (PixelPacket *) NULL)
+    return(DestroyImage(image));
   return(GetFirstImageInList(image));
 }
 
@@ -552,22 +668,23 @@ ModuleExport size_t RegisterTXTImage(void)
   entry->raw=MagickTrue;
   entry->endian_support=MagickTrue;
   entry->description=ConstantString("Sparse Color");
-  entry->module=ConstantString("TXT");
+  entry->magick_module=ConstantString("TXT");
   (void) RegisterMagickInfo(entry);
   entry=SetMagickInfo("TEXT");
-  entry->decoder=(DecodeImageHandler *) ReadTXTImage;
+  entry->decoder=(DecodeImageHandler *) ReadTEXTImage;
   entry->encoder=(EncodeImageHandler *) WriteTXTImage;
   entry->raw=MagickTrue;
   entry->endian_support=MagickTrue;
+  entry->format_type=ImplicitFormatType;
   entry->description=ConstantString("Text");
-  entry->module=ConstantString("TXT");
+  entry->magick_module=ConstantString("TXT");
   (void) RegisterMagickInfo(entry);
   entry=SetMagickInfo("TXT");
   entry->decoder=(DecodeImageHandler *) ReadTXTImage;
   entry->encoder=(EncodeImageHandler *) WriteTXTImage;
   entry->description=ConstantString("Text");
   entry->magick=(IsImageFormatHandler *) IsTXT;
-  entry->module=ConstantString("TXT");
+  entry->magick_module=ConstantString("TXT");
   (void) RegisterMagickInfo(entry);
   return(MagickImageCoderSignature);
 }
@@ -646,6 +763,9 @@ static MagickBooleanType WriteTXTImage(const ImageInfo *image_info,Image *image)
   register ssize_t
     x;
 
+  size_t
+    imageListLength;
+
   ssize_t
     y;
 
@@ -653,29 +773,46 @@ static MagickBooleanType WriteTXTImage(const ImageInfo *image_info,Image *image)
     Open output image file.
   */
   assert(image_info != (const ImageInfo *) NULL);
-  assert(image_info->signature == MagickSignature);
+  assert(image_info->signature == MagickCoreSignature);
   assert(image != (Image *) NULL);
-  assert(image->signature == MagickSignature);
+  assert(image->signature == MagickCoreSignature);
   if (image->debug != MagickFalse)
     (void) LogMagickEvent(TraceEvent,GetMagickModule(),"%s",image->filename);
   status=OpenBlob(image_info,image,WriteBlobMode,&image->exception);
   if (status == MagickFalse)
     return(status);
   scene=0;
+  imageListLength=GetImageListLength(image);
   do
   {
+    ComplianceType
+      compliance;
+
+    const char
+      *value;
+
     (void) CopyMagickString(colorspace,CommandOptionToMnemonic(
       MagickColorspaceOptions,(ssize_t) image->colorspace),MaxTextExtent);
     LocaleLower(colorspace);
     image->depth=GetImageQuantumDepth(image,MagickTrue);
     if (image->matte != MagickFalse)
       (void) ConcatenateMagickString(colorspace,"a",MaxTextExtent);
+    compliance=NoCompliance;
+    value=GetImageOption(image_info,"txt:compliance");
+    if (value != (char *) NULL)
+      compliance=(ComplianceType) ParseCommandOption(MagickComplianceOptions,
+        MagickFalse,value);
     if (LocaleCompare(image_info->magick,"SPARSE-COLOR") != 0)
       {
+        size_t
+          depth;
+
+        depth=compliance == SVGCompliance ? image->depth :
+          MAGICKCORE_QUANTUM_DEPTH;
         (void) FormatLocaleString(buffer,MaxTextExtent,
           "# ImageMagick pixel enumeration: %.20g,%.20g,%.20g,%s\n",(double)
-          image->columns,(double) image->rows,(double)
-          ((MagickOffsetType) GetQuantumRange(image->depth)),colorspace);
+          image->columns,(double) image->rows,(double) GetQuantumRange(depth),
+          colorspace);
         (void) WriteBlobString(image,buffer);
       }
     GetMagickPixelPacket(image,&pixel);
@@ -688,21 +825,15 @@ static MagickBooleanType WriteTXTImage(const ImageInfo *image_info,Image *image)
       for (x=0; x < (ssize_t) image->columns; x++)
       {
         SetMagickPixelPacket(image,p,indexes+x,&pixel);
-        if (pixel.colorspace == LabColorspace)
-          {
-            pixel.green-=(QuantumRange+1)/2.0;
-            pixel.blue-=(QuantumRange+1)/2.0;
-          }
         if (LocaleCompare(image_info->magick,"SPARSE-COLOR") == 0)
           {
             /*
               Sparse-color format.
             */
-            if (GetPixelOpacity(p) == (Quantum) OpaqueOpacity)
+            if ((image->matte == MagickFalse) ||
+                (GetPixelOpacity(p) == (Quantum) OpaqueOpacity))
               {
                 GetColorTuple(&pixel,MagickFalse,tuple);
-                (void) QueryMagickColorname(image,&pixel,SVGCompliance,tuple,
-                  &image->exception);
                 (void) FormatLocaleString(buffer,MaxTextExtent,"%.20g,%.20g,",
                   (double) x,(double) y);
                 (void) WriteBlobString(image,buffer);
@@ -716,20 +847,20 @@ static MagickBooleanType WriteTXTImage(const ImageInfo *image_info,Image *image)
           x,(double) y);
         (void) WriteBlobString(image,buffer);
         (void) CopyMagickString(tuple,"(",MaxTextExtent);
-        ConcatenateColorComponent(&pixel,RedChannel,NoCompliance,tuple);
+        ConcatenateColorComponent(&pixel,RedChannel,compliance,tuple);
         (void) ConcatenateMagickString(tuple,",",MaxTextExtent);
-        ConcatenateColorComponent(&pixel,GreenChannel,NoCompliance,tuple);
+        ConcatenateColorComponent(&pixel,GreenChannel,compliance,tuple);
         (void) ConcatenateMagickString(tuple,",",MaxTextExtent);
-        ConcatenateColorComponent(&pixel,BlueChannel,NoCompliance,tuple);
+        ConcatenateColorComponent(&pixel,BlueChannel,compliance,tuple);
         if (pixel.colorspace == CMYKColorspace)
           {
             (void) ConcatenateMagickString(tuple,",",MaxTextExtent);
-            ConcatenateColorComponent(&pixel,IndexChannel,NoCompliance,tuple);
+            ConcatenateColorComponent(&pixel,IndexChannel,compliance,tuple);
           }
         if (pixel.matte != MagickFalse)
           {
             (void) ConcatenateMagickString(tuple,",",MaxTextExtent);
-            ConcatenateColorComponent(&pixel,AlphaChannel,NoCompliance,tuple);
+            ConcatenateColorComponent(&pixel,AlphaChannel,compliance,tuple);
           }
         (void) ConcatenateMagickString(tuple,")",MaxTextExtent);
         (void) WriteBlobString(image,tuple);
@@ -752,8 +883,7 @@ static MagickBooleanType WriteTXTImage(const ImageInfo *image_info,Image *image)
     if (GetNextImageInList(image) == (Image *) NULL)
       break;
     image=SyncNextImageInList(image);
-    status=SetImageProgress(image,SaveImagesTag,scene++,
-      GetImageListLength(image));
+    status=SetImageProgress(image,SaveImagesTag,scene++,imageListLength);
     if (status == MagickFalse)
       break;
   } while (image_info->adjoin != MagickFalse);
